#include <stdlib.h>
#include <syslog.h>
#include <stdbool.h>
#include <ftdi.h>

#include <gpio.h>

struct ftdi_obj {
    struct ftdi_context *ftdi;
    bool is_open;
    unsigned int vendor_id;
    unsigned int product_id;
    unsigned char bit_mask;     /* Mask of used bits */
    unsigned char status_mask;  /* Mask of current status of used bits */
};

static struct ftdi_obj ftdi_obj = {
    .ftdi = NULL,
    .is_open = false,
    .vendor_id = 0x0403,
    .product_id = 0x6001,
    .bit_mask = 0,
    .status_mask = 0,
};

static int ftdi_open(void)
{
    int res;

    if((ftdi_obj.ftdi = ftdi_new()) == 0) {
        syslog(LOG_ERR, "ftdi_new failed\n");
        return EXIT_FAILURE;
    }

    res = ftdi_usb_open(ftdi_obj.ftdi, ftdi_obj.vendor_id, ftdi_obj.product_id);
    if (res < 0 && res != -5) {
        syslog(LOG_ERR, "unable to open ftdi device: %d (%s)\n", res,
               ftdi_get_error_string(ftdi_obj.ftdi));
        ftdi_free(ftdi_obj.ftdi);
        return EXIT_FAILURE;
    }

    res = ftdi_read_data(ftdi_obj.ftdi, &(ftdi_obj.status_mask), 1);
    if(res < 0) {
        syslog(LOG_ERR, "unable to readfrom ftdi device: %d (%s)\n", res,
               ftdi_get_error_string(ftdi_obj.ftdi));
        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}

int gpio_open(const struct gpio *gpio)
{
    int res;
    unsigned int init_value = 0;

    if(NULL == gpio) {
        syslog(LOG_ERR, "No valid gpio object given.\n");
        return EXIT_FAILURE;
    }

    if(!ftdi_obj.is_open) {
        res = ftdi_open();
        if(res != EXIT_SUCCESS) {
            return EXIT_FAILURE;
        }
    }

    ftdi_obj.bit_mask |= (unsigned char)gpio->pin;
    syslog(LOG_DEBUG, "bitmask: 0x%02x\n", ftdi_obj.bit_mask);
    res = ftdi_set_bitmode(ftdi_obj.ftdi, ftdi_obj.bit_mask, BITMODE_BITBANG);
    if(res < 0) {
        syslog(LOG_ERR, "unable to set bit bang mode: %d (%s)\n", res,
               ftdi_get_error_string(ftdi_obj.ftdi));
        ftdi_usb_close(ftdi_obj.ftdi);
        ftdi_free(ftdi_obj.ftdi);
        return EXIT_FAILURE;
    }

    res = gpio_write(gpio, init_value);
    if(res < 0) {
        syslog(LOG_ERR, "unable to write to gpio %u (%s)\n", gpio->pin,
               ftdi_get_error_string(ftdi_obj.ftdi));
        ftdi_usb_close(ftdi_obj.ftdi);
        ftdi_free(ftdi_obj.ftdi);
        return EXIT_FAILURE;
    }
    ftdi_obj.is_open = true;

    return EXIT_SUCCESS;
}

int gpio_close(const struct gpio *gpio)
{
    if(NULL == gpio) {
        return EXIT_FAILURE;
    }

    gpio_write(gpio, 0);

    ftdi_obj.bit_mask &= ~(unsigned char)(gpio->pin);
    if(ftdi_obj.bit_mask == 0) {
        ftdi_usb_close(ftdi_obj.ftdi);
        ftdi_free(ftdi_obj.ftdi);
    }

    return EXIT_SUCCESS;
}

int gpio_read(const struct gpio *gpio, unsigned int *value)
{
    if(NULL == gpio || NULL == value) {
        return EXIT_FAILURE;
    }

    if(ftdi_obj.status_mask & (unsigned char)(gpio->pin)) {
        *value = 1;
    } else {
        *value = 0;
    }

    return EXIT_SUCCESS;
}

int gpio_write(const struct gpio *gpio, unsigned int value)
{
    int res;
    unsigned char mask;

    if(NULL == gpio) {
        return EXIT_FAILURE;
    }

    if(value == 0) {
        mask &= ~(unsigned char)(gpio->pin);
    }
    else {
        mask |= (unsigned char)(gpio->pin);
    }

    res = ftdi_write_data(ftdi_obj.ftdi, &mask, 1);
    if(res < 0) {
        syslog(LOG_ERR, "write failed for 0x%x, error %d (%s)\n", value, res,
               ftdi_get_error_string(ftdi_obj.ftdi));
        return EXIT_FAILURE;
    }
    ftdi_obj.status_mask = mask;

    return EXIT_SUCCESS;
}

int gpio_toggle(const struct gpio *gpio)
{
    int res;
    unsigned char mask = ftdi_obj.status_mask;

    if(NULL == gpio) {
        return EXIT_FAILURE;
    }

    mask ^= (unsigned char)(gpio->pin);

    res = ftdi_write_data(ftdi_obj.ftdi, &mask, 1);
    if(res < 0) {
        syslog(LOG_ERR, "toggle failed, error %d (%s)\n", res,
               ftdi_get_error_string(ftdi_obj.ftdi));
        return EXIT_FAILURE;
    }
    ftdi_obj.status_mask = mask;

    return EXIT_SUCCESS;
}
